@@ -8,23 +8,36 @@
 #include <QRegularExpression>
 #include <QDebug>
 #include "text.h"
-#include "color.h"
 #include "font.h"
 #include "tile.h"
 #include "pbf.h"
 #include "style.h"
 
 
-static vector_tile::Tile_GeomType geometryType(const QString &str)
+static Data::GeomType geometryType(const QString &str)
 {
 	if (str == "Point")
-		return vector_tile::Tile_GeomType_POINT;
+		return Data::GeomType::POINT;
 	else if (str == "LineString")
-		return vector_tile::Tile_GeomType_LINESTRING;
+		return Data::GeomType::LINESTRING;
 	else if (str == "Polygon")
-		return vector_tile::Tile_GeomType_POLYGON;
+		return Data::GeomType::POLYGON;
 	else
-		return vector_tile::Tile_GeomType_UNKNOWN;
+		return Data::GeomType::UNKNOWN;
+}
+
+static QVariant variant(const QJsonValue &val)
+{
+	switch (val.type()) {
+		case QJsonValue::String:
+			return QVariant(val.toString().toUtf8());
+		case QJsonValue::Double:
+		case QJsonValue::Bool:
+			return val.toVariant();
+		default:
+			qWarning() << val << ": invalid filter value";
+			return QVariant();
+	}
 }
 
 Style::Layer::Filter::Filter(const QJsonArray &json)
@@ -36,50 +49,50 @@ Style::Layer::Filter::Filter(const QJsonArray &json)
 	if (json.isEmpty())
 		INVALID_FILTER(json);
 
-	QString type = json.at(0).toString();
+	QString type(json.at(0).toString());
 
 	if (type == "==") {
 		if (json.size() != 3)
 			INVALID_FILTER(json);
 		if (json.at(1).toString() == "$type") {
 			_type = GeometryType;
-			_kv = QPair<QString, QVariant>(QString(),
+			_kv = QPair<QByteArray, QVariant>(QByteArray(),
 			  QVariant(geometryType(json.at(2).toString())));
 		} else {
 			_type = EQ;
-			_kv = QPair<QString, QVariant>(json.at(1).toString(),
-			  json.at(2).toVariant());
+			_kv = QPair<QByteArray, QVariant>(json.at(1).toString().toUtf8(),
+			  variant(json.at(2)));
 		}
 	} else if (type == "!=") {
 		if (json.size() != 3)
 			INVALID_FILTER(json);
 		_type = NE;
-		_kv = QPair<QString, QVariant>(json.at(1).toString(),
-		  json.at(2).toVariant());
+		_kv = QPair<QByteArray, QVariant>(json.at(1).toString().toUtf8(),
+		  variant(json.at(2)));
 	} else if (type == "<") {
 		if (json.size() != 3)
 			INVALID_FILTER(json);
 		_type = LT;
-		_kv = QPair<QString, QVariant>(json.at(1).toString(),
-		  json.at(2).toVariant());
+		_kv = QPair<QByteArray, QVariant>(json.at(1).toString().toUtf8(),
+		  variant(json.at(2)));
 	} else if (type == "<=") {
 		if (json.size() != 3)
 			INVALID_FILTER(json);
 		_type = LE;
-		_kv = QPair<QString, QVariant>(json.at(1).toString(),
-		  json.at(2).toVariant());
+		_kv = QPair<QByteArray, QVariant>(json.at(1).toString().toUtf8(),
+		  variant(json.at(2)));
 	} else if (type == ">") {
 		if (json.size() != 3)
 			INVALID_FILTER(json);
 		_type = GT;
-		_kv = QPair<QString, QVariant>(json.at(1).toString(),
-		  json.at(2).toVariant());
+		_kv = QPair<QByteArray, QVariant>(json.at(1).toString().toUtf8(),
+		  variant(json.at(2)));
 	} else if (type == ">=") {
 		if (json.size() != 3)
 			INVALID_FILTER(json);
 		_type = GE;
-		_kv = QPair<QString, QVariant>(json.at(1).toString(),
-		  json.at(2).toVariant());
+		_kv = QPair<QByteArray, QVariant>(json.at(1).toString().toUtf8(),
+		  variant(json.at(2)));
 	} else if (type == "all") {
 		_type = All;
 		for (int i = 1; i < json.size(); i++)
@@ -92,28 +105,32 @@ Style::Layer::Filter::Filter(const QJsonArray &json)
 		if (json.size() < 3)
 			INVALID_FILTER(json);
 		_type = In;
-		_kv = QPair<QString, QVariant>(json.at(1).toString(), QVariant());
+		_kv = QPair<QByteArray, QVariant>(json.at(1).toString().toUtf8(),
+		  QVariant());
 		for (int i = 2; i < json.size(); i++)
-			_set.insert(json.at(i).toString());
+			_set.insert(json.at(i).toString().toUtf8());
 	}  else if (type == "!in") {
 		if (json.size() < 3)
 			INVALID_FILTER(json);
 		_type = In;
 		_not = true;
-		_kv = QPair<QString, QVariant>(json.at(1).toString(), QVariant());
+		_kv = QPair<QByteArray, QVariant>(json.at(1).toString().toUtf8(),
+		  QVariant());
 		for (int i = 2; i < json.size(); i++)
-			_set.insert(json.at(i).toString());
+			_set.insert(json.at(i).toString().toUtf8());
 	} else if (type == "has") {
 		if (json.size() < 2)
 			INVALID_FILTER(json);
 		_type = Has;
-		_kv = QPair<QString, QVariant>(json.at(1).toString(), QVariant());
+		_kv = QPair<QByteArray, QVariant>(json.at(1).toString().toUtf8(),
+		  QVariant());
 	} else if (type == "!has") {
 		if (json.size() < 2)
 			INVALID_FILTER(json);
 		_type = Has;
 		_not = true;
-		_kv = QPair<QString, QVariant>(json.at(1).toString(), QVariant());
+		_kv = QPair<QByteArray, QVariant>(json.at(1).toString().toUtf8(),
+		  QVariant());
 	} else
 		INVALID_FILTER(json);
 }
@@ -183,7 +200,7 @@ bool Style::Layer::Filter::match(const PBF::Feature &feature) const
 			if (!(v = feature.value(_kv.first)))
 				return _not;
 			else
-				return _set.contains((*v).toString()) ^ _not;
+				return _set.contains((*v).toByteArray()) ^ _not;
 		case Has:
 			return (feature.value(_kv.first) ? true : false) ^ _not;
 		case All:
@@ -205,7 +222,7 @@ bool Style::Layer::Filter::match(const PBF::Feature &feature) const
 
 QString Style::Layer::Template::value(int zoom, const PBF::Feature &feature) const
 {
-	QRegularExpression rx("\\{[^\\}]*\\}");
+	static QRegularExpression rx("\\{[^\\}]*\\}");
 	QString text(_field.value(zoom));
 	QRegularExpressionMatchIterator it = rx.globalMatch(text);
 	QStringList keys;
@@ -217,7 +234,7 @@ QString Style::Layer::Template::value(int zoom, const PBF::Feature &feature) con
 	}
 	for (int i = 0; i < keys.size(); i++) {
 		const QString &key = keys.at(i);
-		const QVariant *val = feature.value(key);
+		const QVariant *val = feature.value(key.toUtf8());
 		text.replace(QString("{%1}").arg(key), val ? val->toString() : "");
 	}
 
@@ -258,6 +275,9 @@ Style::Layer::Paint::Paint(const QJsonObject &json)
 	_textHaloColor = FunctionC(json["text-halo-color"], QColor());
 	_textHaloWidth = FunctionF(json["text-halo-width"]);
 	_textHaloBlur = FunctionF(json["text-halo-blur"]);
+
+	// icon
+	_iconColor = FunctionC(json["icon-color"]);
 }
 
 QPen Style::Layer::Paint::pen(Type type, int zoom) const
@@ -293,7 +313,7 @@ QPen Style::Layer::Paint::pen(Type type, int zoom) const
 	return pen;
 }
 
-QBrush Style::Layer::Paint::brush(Type type, int zoom, const Sprites &sprites)
+QBrush Style::Layer::Paint::brush(Type type, int zoom, Sprites &sprites)
   const
 {
 	QColor color;
@@ -370,6 +390,7 @@ Style::Layer::Layout::Layout(const QJsonObject &json)
 
 	// icon
 	_icon = Template(FunctionS(json["icon-image"]));
+	_iconSize = FunctionF(json["icon-size"]);
 
 	// symbol
 	_symbolPlacement = FunctionS(json["symbol-placement"]);
@@ -394,9 +415,10 @@ Text::Anchor Style::Layer::Layout::textAnchor(int zoom) const
 {
 	QString anchor(_textAnchor.value(zoom));
 
-	if (anchor == "left")
+	if (anchor == "left" || anchor == "top-left" || anchor == "bottom-left")
 		return Text::Left;
-	else if (anchor == "right")
+	else if (anchor == "right" || anchor == "top-right"
+	  || anchor == "bottom-right")
 		return Text::Right;
 	else if (anchor == "top")
 		return Text::Top;
@@ -483,7 +505,7 @@ Style::Layer::Layer(const QJsonObject &json)
 		_type = Symbol;
 
 	// source-layer
-	_sourceLayer = json["source-layer"].toString();
+	_sourceLayer = json["source-layer"].toString().toUtf8();
 
 	// zooms
 	if (json.contains("minzoom") && json["minzoom"].isDouble())
@@ -512,7 +534,7 @@ bool Style::Layer::match(int zoom, const PBF::Feature &feature) const
 	return _filter.match(feature);
 }
 
-void Style::Layer::setPathPainter(Tile &tile, const Sprites &sprites) const
+void Style::Layer::setPathPainter(Tile &tile, Sprites &sprites) const
 {
 	QPainter &p = tile.painter();
 	int zoom = tile.zoom();
@@ -545,14 +567,18 @@ void Style::Layer::setTextProperties(Tile &tile) const
 }
 
 void Style::Layer::addSymbol(Tile &tile, const QPainterPath &path,
-  const PBF::Feature &feature, const Sprites &sprites) const
+  const PBF::Feature &feature, Sprites &sprites) const
 {
-	QString text = _layout.text(tile.zoom(), feature);
-	if (text.isEmpty())
+	QString text(_layout.text(tile.zoom(), feature));
+	QString icon(_layout.icon(tile.zoom(), feature));
+	QColor color(_paint.iconColor(tile.zoom()));
+	qreal size(_layout.iconSize(tile.zoom()));
+	QImage img(sprites.icon(icon, color, size));
+
+	if (text.isEmpty() && img.isNull())
 		return;
 
-	QString icon = _layout.icon(tile.zoom(), feature);
-	tile.text().addLabel(text, sprites.icon(icon), path);
+	tile.text().addLabel(text, img, path);
 }
 
 static bool loadSprites(const QDir &styleDir, const QString &json,
@@ -565,7 +591,7 @@ static bool loadSprites(const QDir &styleDir, const QString &json,
 		if (QFileInfo::exists(spritesImg))
 			return sprites.load(spritesJSON, spritesImg);
 		else {
-			qCritical() << spritesImg << ": no such file";
+			qWarning() << spritesImg << ": no such file";
 			return false;
 		}
 	}
@@ -598,20 +624,20 @@ bool Style::load(const QString &fileName)
 				_layers.append(Layer(layers[i].toObject()));
 	}
 
-	QDir styleDir = QFileInfo(fileName).absoluteDir();
+	QDir styleDir(QFileInfo(fileName).absoluteDir());
 	loadSprites(styleDir, "sprite.json", "sprite.png", _sprites);
 	loadSprites(styleDir, "sprite@2x.json", "sprite@2x.png", _sprites2x);
 
 	return true;
 }
 
-const Sprites &Style::sprites(const QPointF &scale) const
+Sprites &Style::sprites(const QPointF &scale)
 {
 	return (scale.x() > 1.0 || scale.y() > 1.0)
 	  && !_sprites2x.isNull() ? _sprites2x : _sprites;
 }
 
-void Style::setupLayer(Tile &tile, const Layer &layer) const
+void Style::setupLayer(Tile &tile, const Layer &layer)
 {
 	if (layer.isSymbol())
 		layer.setTextProperties(tile);
@@ -619,7 +645,7 @@ void Style::setupLayer(Tile &tile, const Layer &layer) const
 		layer.setPathPainter(tile, sprites(tile.scale()));
 }
 
-void Style::drawBackground(Tile &tile) const
+void Style::drawBackground(Tile &tile)
 {
 	QRectF rect(QPointF(0, 0), QSizeF(tile.size().width() / tile.scale().x(),
 	  tile.size().height() / tile.scale().y()));
@@ -634,13 +660,10 @@ void Style::drawBackground(Tile &tile) const
 		_layers.first().setPathPainter(tile, _sprites);
 		tile.painter().drawPath(path);
 	}
-
-	//tile.painter().setPen(Qt::red);
-	//tile.painter().drawRect(rect);
 }
 
 void Style::drawFeature(const PBF::Feature &feature, const Layer &layer,
-  Tile &tile, const QSizeF &factor) const
+  Tile &tile, const QSizeF &factor)
 {
 	if (!layer.match(tile.zoom(), feature))
 		return;
@@ -656,17 +679,17 @@ void Style::drawFeature(const PBF::Feature &feature, const Layer &layer,
 }
 
 void Style::drawLayer(const PBF::Layer &pbfLayer, const Layer &styleLayer,
-  Tile &tile) const
+  Tile &tile)
 {
-	if (pbfLayer.data()->version() > 2)
+	if (pbfLayer.data()->version > 2)
 		return;
 
 	if (!styleLayer.isVisible())
 		return;
 
 	QSizeF factor(tile.size().width() / tile.scale().x() /
-	  (qreal)pbfLayer.data()->extent(), tile.size().height() / tile.scale().y()
-	  / (qreal)pbfLayer.data()->extent());
+	  (qreal)pbfLayer.data()->extent, tile.size().height() / tile.scale().y()
+	  / (qreal)pbfLayer.data()->extent);
 
 	tile.painter().save();
 	setupLayer(tile, styleLayer);
@@ -675,12 +698,12 @@ void Style::drawLayer(const PBF::Layer &pbfLayer, const Layer &styleLayer,
 	tile.painter().restore();
 }
 
-void Style::render(const PBF &data, Tile &tile) const
+void Style::render(const PBF &data, Tile &tile)
 {
 	drawBackground(tile);
 
 	for (int i = 0; i < _layers.size(); i++) {
-		QHash<QString, PBF::Layer*>::const_iterator it = data.layers().find(
+		QHash<QByteArray, PBF::Layer*>::const_iterator it = data.layers().find(
 		  _layers.at(i).sourceLayer());
 		if (it == data.layers().constEnd())
 			continue;
@@ -689,4 +712,11 @@ void Style::render(const PBF &data, Tile &tile) const
 	}
 
 	tile.text().render(&tile.painter());
+
+	//QRectF rect(QPointF(0, 0), QSizeF(tile.size().width() / tile.scale().x(),
+	//  tile.size().height() / tile.scale().y()));
+	//tile.painter().setPen(Qt::red);
+	//tile.painter().setBrush(Qt::NoBrush);
+	//tile.painter().setRenderHint(QPainter::Antialiasing, false);
+	//tile.painter().drawRect(rect);
 }
